%% EXAMPLE: Differential drive vehicle following waypoints using the 
% Pure Pursuit algorithm (robotics.PurePursuit)
%
% Copyright 2018 The MathWorks, Inc.

%% Define Vehicle
R = 0.1;                % Wheel radius [m]
L = 0.5;                % Wheelbase [m]
dd = DifferentialDrive(R,L);

%% Simulation parameters
sampleTime = 0.05;             % Sample time [s]
tVec = 0:sampleTime:35;         % Time array

initPose = [1.5;1.5;0];             % Initial pose (x y theta) %  %answer (iv)-changed from [0;0;0] to [1.5;1.5;0]
pose = zeros(3,numel(tVec));    % Pose matrix
pose(:,1) = initPose;

% Define waypoints
% waypoints = [0,0; 2,2; 4,2; 2,4; 0.5,3]; %commented for answer (iv)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% answer (iv)
%% Path planning
% Load map and inflate it by a safety distance
close all
load exampleMap
inflate(map,R);

% Create a Probabilistic Road Map (PRM)
planner = robotics.PRM(map);
planner.NumNodes = 75;
planner.ConnectionDistance = 5;

% Find a path from the start point to a specified goal point
startPoint = initPose(1:2)';
goalPoint  = [11, 2]; 
waypoints = findpath(planner,startPoint,goalPoint);
show(planner)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Create visualizer %commented for answer (iv)
% viz = Visualizer2D;
% viz.hasWaypoints = true;

load exampleMap % Reload original (uninflated) map for visualization %added for answer (iv)
viz = Visualizer2D;
viz.hasWaypoints = true;
viz.mapName = 'map';

%% Pure Pursuit Controller
controller = robotics.PurePursuit;
controller.Waypoints = waypoints;
controller.LookaheadDistance = 0.35;
controller.DesiredLinearVelocity = 0.75;
controller.MaxAngularVelocity = 1.5;

%% Simulation loop
% close all  %commented for answer (iv)
r = robotics.Rate(1/sampleTime);
count =0;

for idx = 2:numel(tVec) 
    % Run the Pure Pursuit controller and convert output to wheel speeds
    [vRef,wRef] = controller(pose(:,idx-1));
    [wL,wR] = inverseKinematics(dd,vRef,wRef);
    
    % Compute the velocities
    [v,w] = forwardKinematics(dd,wL,wR);
    velB = [v;0;w]; % Body velocities [vx;vy;w]
    vel = bodyToWorld(velB,pose(:,idx-1));  % Convert from body to world
    
    % Perform forward discrete integration step
    pose(:,idx) = pose(:,idx-1) + vel*sampleTime; 
    
    % Update visualization
    viz(pose(:,idx),waypoints)
    waitfor(r);
    

% answer for (ii)
%     if idx==100
%         waypoints = [0,0; 2,2; 4,2; 0.5,3];
%         controller.Waypoints = waypoints;
%     end

% answer for (i)        
    dist_to_lastWP = sqrt((waypoints(end,1)-pose(1,idx))^2 + (waypoints(end,2)-pose(2,idx))^2);
    if dist_to_lastWP<0.1
        break
    end
    
% answer for (iii)
%     if count<5
%         waypoints(end,2) = waypoints(end,2) + 0.2;
%         controller.Waypoints = waypoints;
%         count = count + 1;
%     end
%     if count>=5
%         waypoints(end,2) = waypoints(end,2) - 0.2;
%         controller.Waypoints = waypoints;
%         count = count + 1;
%         if count==10
%             count = 0;
%         end        
%     end
    
end